@@ -270,6 +270,33 @@ protected:
     if (this->Opts.HaveCC) {
       builtins += this->Opts.Predefines;
 
+      // Remove GCC builtin definitions for features Clang does not implement.
+      if (this->IsActualGNU(this->Opts.Predefines)) {
+        builtins += "#undef __BFLT16_DECIMAL_DIG__\n"
+                    "#undef __BFLT16_DENORM_MIN__\n"
+                    "#undef __BFLT16_DIG__\n"
+                    "#undef __BFLT16_DIG__\n"
+                    "#undef __BFLT16_EPSILON__\n"
+                    "#undef __BFLT16_HAS_DENORM__\n"
+                    "#undef __BFLT16_HAS_INFINITY__\n"
+                    "#undef __BFLT16_HAS_QUIET_NAN__\n"
+                    "#undef __BFLT16_IS_IEC_60559__\n"
+                    "#undef __BFLT16_MANT_DIG__\n"
+                    "#undef __BFLT16_MAX_10_EXP__\n"
+                    "#undef __BFLT16_MAX_EXP__\n"
+                    "#undef __BFLT16_MAX__\n"
+                    "#undef __BFLT16_MIN_10_EXP__\n"
+                    "#undef __BFLT16_MIN_EXP__\n"
+                    "#undef __BFLT16_MIN__\n"
+                    "#undef __BFLT16_NORM_MAX__\n"
+                    "#undef __SIZEOF_FLOAT80__\n"
+                    "#undef __STDCPP_BFLOAT16_T__\n"
+                    "#undef __STDCPP_FLOAT128_T__\n"
+                    "#undef __STDCPP_FLOAT16_T__\n"
+                    "#undef __STDCPP_FLOAT32_T__\n"
+                    "#undef __STDCPP_FLOAT64_T__\n";
+      }
+
       // Provide __builtin_va_arg_pack if simulating the actual GNU compiler.
       if (this->NeedBuiltinVarArgPack(this->Opts.Predefines)) {
         // Clang does not support this builtin, so fake it to tolerate
@@ -380,6 +407,12 @@ protected:
         builtins += "\n"
                     "#define __malloc__(...) __malloc__\n";
       }
+      if (this->NeedAttributeAssumeSuppression(this->Opts.Predefines)) {
+        // Clang does not support '__attribute__((__assume__(args...)))'
+        // as a statement attribute used in libstdc++ headers.
+        builtins += "\n"
+                    "#define __assume__(...)\n";
+      }
 
       // Clang's arm_neon.h checks for a feature macro not defined by GCC.
       if (this->NeedARMv8Intrinsics(this->Opts.Predefines)) {
@@ -392,26 +425,57 @@ protected:
         // Clang does not have these types for all sizes.
         // Provide our own approximation of the builtins.
         builtins += "\n"
-                    "typedef struct __castxml_Float32_s { "
-                    "  char x[4] __attribute__((aligned(4))); "
-                    "} __castxml_Float32;\n"
                     "#define _Float32 __castxml_Float32\n"
-                    "typedef struct __castxml_Float32x_s { "
-                    "  char x[8] __attribute__((aligned(8))); "
-                    "} __castxml_Float32x;\n"
                     "#define _Float32x __castxml_Float32x\n"
-                    "typedef struct __castxml_Float64_s { "
-                    "  char x[8] __attribute__((aligned(8))); "
-                    "} __castxml_Float64;\n"
                     "#define _Float64 __castxml_Float64\n"
-                    "typedef struct __castxml_Float64x_s { "
-                    "  char x[16] __attribute__((aligned(16))); "
-                    "} __castxml_Float64x;\n"
-                    "#define _Float64x __castxml_Float64x\n"
-                    "typedef struct __castxml_Float128_s { "
-                    "  char x[16] __attribute__((aligned(16))); "
-                    "} __castxml_Float128;\n"
-                    "#define _Float128 __castxml_Float128\n";
+                    "#define _Float64x __castxml_Float64x\n";
+        if (this->NeedFloat128(this->Opts.Predefines)) {
+          builtins += "#define _Float128 __castxml_Float128\n";
+        }
+
+        if (this->IsCPlusPlus(this->Opts.Predefines)) {
+          // In C++ we need distinct types for template specializations
+          // in glibc headers, but also need conversions.
+          builtins += "\n"
+                      "typedef struct __castxml_Float32_s { "
+                      "  float x; "
+                      "  operator float() const; "
+                      "  __castxml_Float32_s(float); "
+                      "} __castxml_Float32;\n"
+                      "typedef struct __castxml_Float32x_s { "
+                      "  double x; "
+                      "  operator double() const; "
+                      "  __castxml_Float32x_s(double); "
+                      "} __castxml_Float32x;\n"
+                      "typedef struct __castxml_Float64_s { "
+                      "  double x; "
+                      "  operator double() const; "
+                      "  __castxml_Float64_s(double); "
+                      "} __castxml_Float64;\n"
+                      "typedef struct __castxml_Float64x_s { "
+                      "  long double x; "
+                      "  operator long double() const; "
+                      "  __castxml_Float64x_s(long double); "
+                      "} __castxml_Float64x;\n";
+          if (this->NeedFloat128(this->Opts.Predefines)) {
+            builtins += "typedef struct __castxml_Float128_s { "
+                        "  __float128 x; "
+                        "  operator __float128() const; "
+                        "  __castxml_Float128_s(__float128); "
+                        "} __castxml_Float128;\n";
+          }
+
+        } else {
+          // In C we need real float types for conversions in glibc headers.
+          builtins += "\n"
+                      "typedef float __castxml_Float32;\n"
+                      "typedef double __castxml_Float32x;\n"
+                      "typedef double __castxml_Float64;\n"
+                      "typedef long double __castxml_Float64x;\n";
+          if (this->NeedFloat128(this->Opts.Predefines)) {
+            builtins += "typedef __float128 __castxml_Float128;\n";
+          }
+        }
       }
 
     } else {
@@ -463,6 +527,11 @@ protected:
     return this->IsActualGNU(pd);
   }
 
+  bool NeedAttributeAssumeSuppression(std::string const& pd)
+  {
+    return this->IsActualGNU(pd);
+  }
+
   bool NeedFloat128(std::string const& pd) const
   {
     return (this->IsActualGNU(pd) &&
